@@ -3,7 +3,7 @@
 # Main SMTP protocol implementation
 #
 # This file is part of SAUCE, a very picky anti-spam receiver-SMTP.
-# SAUCE is Copyright (C) 1997-1999 Ian Jackson
+# SAUCE is Copyright (C) 1997-2001 Ian Jackson
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -19,7 +19,7 @@
 # along with this program; if not, write to the Free Software Foundation,
 # Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 #
-# $Id: smtp.tcl,v 1.1 1999/10/09 15:46:15 ian Exp $
+# $Id: smtp.tcl,v 1.6 2001/03/21 23:53:15 ian Exp $
 
 
 ########## connection threads
@@ -28,67 +28,91 @@
 # errors/results ignored
 #
 # state variables:
+#
+# Always set:
 # chan              incoming SMTP connection
 # ra                calling IP address
 # rp                calling port
 # rh                calling hostname (or IP-literal), available only after HELO
-# cmdomain          calling host's mail domain to report to
 # la                called (local) address
 # lh                called (local) canonical hostname for interface
 # lp                called (local) port
-# ident             result of ident lookup (whether informative or error)
+# rcptcounter       counter of total no of RCPTs issued in session, for logging
+# smtpcmd           SMTP command we are processing, or empty string
+# smtpreaderr       Last system error reading from incoming SMTP connection
+# rej_conn          list of connection-wide rejection reasons (or {})
+# defer_conn        deferral reasons (defer_* takes precedence over rej_*)
+#                    none of these reasons may contain "\n"
+# smtperrors        count of SMTP error responses
+# quitting          set => we have been asked to shut down
+
+# Generally useful
+# avfid             ID of current address verification thread (unset => none)
 # dnsid             thread for forward DNS lookups (unset => none)
 # dnsptrid          thread for reverse DNS lookup (unset => none)
+# lastchal          last challenge sent on this conn for SAUCEADMIN
+# rbl_hits          list of RBL domains we've hit
+
+# Used during connection setup, but not changed afterwards:
+# cmdomain          calling host's mail domain to report to
+# ident             result of ident lookup (whether informative or error)
 # ichan             reverse ident lookup connection (unset => none)
-# itoid             timeout for ident lookup or incoming command (unset => none)
-# mtachan           channel to local MTA
-# avfid             ID of current address verification thread (unset => none)
-# quitting          set => we have been asked to shut down
-# smtperrors        count of SMTP error responses
-#
-# state variables to do with incoming SMTP state machine:
+# itoid             timeout for ident lookup or incoming command (unset=>none)
+
+# Only set after HELO
 # helostring        set => we have had helo/ehlo, and this was what they said
 #                   this implies mtachan being open and connected
-# helocmd           the command helo or ehlo that they used
 # smtpstyle         `smtp' or `esmtp'.
-# mf_lp             set => we have had MAIL FROM, and this is the local part
-# mf_dm             domain from MAIL FROM, may be garbage or unset if mf_lp not set
-# no                set => we are in one of the `NO' states from the overall
-#                   state machine, and want to reject the incoming mail
-#                   (this means states RN, MN, and associated incoming data areas
-#                   of the flowchart)
-#                   If add_bl is set then we hide this reason from the caller;
-#                   otherwise we show it.
-# defer             set => we want to defer this message due to lack of trust
-#                   (no takes precedence).  This string is the reason string.
-#
-# whyreject         reason we are rejecting (if not in message)
-# add_bl            `add-bl': we want to add something to the blacklist
-# a_admin           `a-a': we have had an admin recipient and said 250
-# a_normal          `a-n': we have had a normal recipient and said 250
-# a_bypass          `a-b': we have had a bypass-check recipient and said 250
-# t_w               `t-w': recipient addresses that we want to add to the
-#                    whitelist if we accept the message (checked originators)
-# lastchal          last challenge sent on this conn for SAUCEADMIN
-# smtpcmd           SMTP command we are processing, or empty string
+# mtachan           channel to local MTA
+# helocmd           the command helo or ehlo that they used
+
+# Only set after MAIL FROM; unset/reset for new messages:
+# rej_msg           list of reasons why we're rejecting this message
+# defer_msg         list of reasons why we're deferring this message
+#                    (only ever set before we get to DATA to avoid late defer)
+# delay_msg         set => we want to defer because of an RBL or newness &c
 # att_rcpts         list of recipients they've asked for (perhaps not got)
-#
-# Recipient deferral:
-# We can defer if we get mail from unknown senders or unknown hosts.
-# The deferral period can be increased if the sending host is RBLd.
-# So: both sender and host on whitelist => no deferral, no RBL lookup.
-# Otherwise, do RBL lookup.  Sending host RBLd for reject => reject.
-# Otherwise, defer until BOTH
-#  first contact from this site was at least minimum site deferral ago
-#  first contact from this sender was at least minimum sender deferral ago
-# (minimum is calculated across all applicable RBLs, and new
-# site or sender deferral).
-# 
+#		     each entry is list: $class $rcpt $respcoode
+# mf_lp             set => we have had MAIL FROM, and this is the local part
+# mf_dm             MAIL FROM domain, may be garbage or unset if mf_lp not set
+# a_kinds           list of recip kinds we've accepted (unchecked/normal/bait)
+#                    we only allow ourselves to get into the following states:
+#                     none unchecked normal bait unchecked+bait normal+bait
+#                     and the same +lax (but avoid accepting normal after bait)
+#                    for this var, nodelay addrs are normal.  Repeats allowed.
 # mf_parms          parameters to MAIL FROM, if any
+
+# add_bl            list of reasons to blacklist everyone to do with this msg
+#                    none may contain newline
+
+# Used/set while processing message body ([re]set by ic_msg_resetvars)
+# header            accumulated header text
+# currenthn         name of header field we're currently in
+# currenthl         value of header field we're currently in, including name
+# mid               value of Message-ID header field
+# resentmids        list of Resent-Message-ID's ({} => none)
+# resentany         1 => we have had a Resent- field, 0 => we haven't
+# hdrscomplete      we have received and stored all the headers (that we
+#                   are going to parse); now finish verifs & do body (unset/1)
+# originators       List of origs we've seen (incl. env & headers) & checked
+# originators_tochk Originators seen but not yet checked (with avf)
+#                    (both in dequoted form), as [list $lp $dm $headername]
+
+# Recipient delay:
+# We can delay if we get mail from unknown senders or unknown hosts.
+# The delay period can be increased if the sending host is RBLd.
+# So: both sender and host on whitelist => no delay, no RBL lookup.
+# Otherwise, do RBL lookup.  Sending host RBLd for reject => reject.
+# Otherwise, delay until BOTH
+#  first contact from this site was at least minimum site delay ago
+#  first contact from this sender was at least minimum sender delay ago
+# (minimum is calculated across all applicable RBLs, and new
+# site or sender delay).
+# 
 # mf_message        proposed success message for MAIL FROM
 # rblids            RBL lookups in progress (or empty list)
-# minsiteage        minimum time (in seconds) since first contact with site, or we defer
-# minaddrage        minimum time (in seconds) since first contact with addr, or we defer
+# minsiteage        min time (secs) since first contact with site, or we delay
+# minaddrage        min time (secs) since first contact with addr, or we delay
 # mf_lp, mf_dm      set
 
 thread_typedefine ic {chan lalhlp ra rp} {
@@ -96,16 +120,27 @@ thread_typedefine ic {chan lalhlp ra rp} {
     set state(chan) $chan
     set state(ra) $ra
     set state(rp) $rp
+    manyset $lalhlp state(la) state(lh) state(lp)
     set state(smtpcmd) {}
     set state(smtpreaderr) {}
-    set state(whyreject) {}
+    set state(rej_conn) {}
+    set state(defer_conn) {}
     set state(smtperrors) 0
+    set state(rcptcounter) 0
+
+    set state(rej_msg) {}
+    set state(defer_msg) {}
+    set state(delay_msg) {}
     set state(att_rcpts) {}
+    set state(rbl_hits) {}
+
     set state(rblids) {}
-    manyset $lalhlp state(la) state(lh) state(lp)
-    set state(dnsptrid) [thread_start dnsptr "$state(desc) / reverse lookup" $ra]
+    set state(resentmids) {}
+    set state(dnsptrid) \
+	    [thread_start dnsptr "$state(desc) / reverse lookup" $ra]
     if {[catch {
-	set state(ichan) [socket -myaddr $state(la) -async $state(ra) $ident_port]
+	set state(ichan) [socket -myaddr $state(la) -async \
+		                 $state(ra) $ident_port]
 	chanset_desc $state(ichan) "$state(desc) / ident"
     } emsg]} {
 	log notice "ident error connecting to $ra: $emsg"
@@ -119,13 +154,19 @@ thread_typedefine ic {chan lalhlp ra rp} {
     global canonical_hostname
     set state(quitting) 1
     if {![info exists state(header)] && ![info exists state(sofar)]} {
+	catch_close_cleardesc state(mtachan)
 	ic_commandfinalresponse immed "421 $canonical_hostname shutting down"
     }
 } {
-    foreach thr $state(rblids) { catch { thread_cancel dns $thr } }
-    set state(rblids) {}
+    ic_kill_subthreads
     catch { fileevent $state(chan) readable {} }
     catch { fileevent $state(chan) writable {} }
+    catch_close_cleardesc state(mtachan)
+}
+
+thread_subproc ic kill_subthreads {} {
+    foreach thr $state(rblids) { catch { thread_cancel dns $thr } }
+    set state(rblids) {}
     catch_close_cleardesc state(ichan)
     catch { thread_cancel dnsptr $state(dnsptrid) }
     catch { thread_cancel avf $state(avfid) }
@@ -133,7 +174,6 @@ thread_typedefine ic {chan lalhlp ra rp} {
     catch { unset state(itoid) }
     catch { after cancel $state(ptoid) }
     catch { unset state(ptoid) }
-    catch_close_cleardesc state(mtachan)
 }
 
 thread_chainproc ic ident_timeout {} {
@@ -162,6 +202,9 @@ thread_chainproc ic ident_rx_ok {data} {
     } elseif {[regexp -nocase {^[ \t]*[0-9]+[ \t]*,[ \t]*[0-9]+[ \t]*:[ \t]*error[ \t]*:(.*)$} $data all error]} {
 	log notice "ident remote error on $state(ra): [string trim $error]"
 	ic_ident_done {}
+    } else {
+	log notice "ident gave garbage on $state(ra): [string trim $data]"
+	ic_ident_done {}
     }
 }
 
@@ -177,29 +220,32 @@ thread_subproc ic ident_done {ident} {
     thread_join ic $id dnsptr $state(dnsptrid) remotedns_ok remotedns_err
 }
 
-thread_chainproc ic remotedns_ok {answers} {
-    unset state(dnsptrid)
+thread_chainproc ic remotedns_ok {answers emsgstr} {
+    global require_reverse_dns
     if {[llength $answers]} {
 	ic_remotedns_done [join $answers]
     } else {
+	if {$require_reverse_dns} {
+	    lappend state(rej_conn) "reverse DNS: $state(ra): $emsgstr"
+	}
 	ic_remotedns_done "\[$state(ra)\]"
     }
 }
 
 thread_chainproc ic remotedns_err {emsg} {
-    global require_reverse_dns canonical_hostname
+    global require_reverse_dns
     if {$require_reverse_dns} {
-	ic_commandfinalresponse major \
-		"421 $canonical_hostname $state(ra): reverse DNS: $emsg"
-    } else {
-	ic_remotedns_done "\[$state(ra)\]"
+	lappend state(defer_conn) "reverse DNS: $state(ra): [singleline $emsg]"
     }
+    ic_remotedns_done "\[$state(ra)\]"
 }
 
 thread_subproc ic remotedns_done {dnsresult} {
     global canonical_hostname
+    unset state(dnsptrid)
     set state(rh) $dnsresult
-    ic_commandresponse greeting "220 $canonical_hostname sauce-smtpd ESMTP ready"
+    ic_commandresponse_maybefinal greeting -1 \
+	    "220 $canonical_hostname sauce-smtpd ESMTP ready"
 }
 
 thread_subproc ic commandresponse {evtype response} {
@@ -251,6 +297,8 @@ thread_subproc ic getsiteannoy {change} {
 }
 
 thread_subproc ic commandresponse_maybefinal {evtype final response} {
+    # final==-1 means initial (used because then we have to prefix
+    # each stalling line); final==0 means normal; final==1 means final.
     global max_smtp_errors canonical_hostname
     global annoy_actout_max annoy_grumpy annoy_partrespevery annoy_actout_nopartresp
     global pleasure_command pleasure_delivery annoyance_major annoyance_minor
@@ -272,7 +320,7 @@ thread_subproc ic commandresponse_maybefinal {evtype final response} {
 	    set delay $cv
 	    incr state(smtperrors)
 
-	    if {!$final && $state(smtperrors) > $max_smtp_errors} {
+	    if {$final<=0 && $state(smtperrors) > $max_smtp_errors} {
 		log notice "too many errors from $state(rh), closing channel (annoy=$cv)"
 		ic_commandfinalresponse major \
 			"421 $canonical_hostname $response \[too many errors\]"
@@ -306,26 +354,30 @@ thread_subproc ic commandresponse_maybefinal {evtype final response} {
 	default { error "$evtype ?" }
     }
 
-    if {[string length $response] && ($delay > 0 || "$evtype" == "greeting")} {
-	regsub {(?m)$} $response " \[$irritamt\]" response
+    if {[string length $response]} {
+	if {$delay > 0 || "$evtype" == "greeting"} {
+	    regsub {(?m)$} $response " \[[irrit_present $irritamt]\]" response
+	}
+    } else {
+	set delay 0
     }
-    ic_commandresponsedelay $delay [expr {"$evtype" == "nopartresp"}] $final $response
+    ic_commandresponsedelay $delay [expr {"$evtype" == "nopartresp"}] \
+	    $final $response 0
 }
 
-thread_subproc ic commandresponsedelay {delay nopartresp final response} {
+thread_subproc ic commandresponsedelay {delay nopartresp final response f_ix} {
     global command_timeout annoy_partrespevery
 
-    if {!$final && [ic_check_quitting]} { return }
+    if {$final<=0 && [ic_check_quitting]} { return }
 
-    if {!$nopartresp && $delay > $annoy_partrespevery && \
-	    [regsub {(?m)^([0-9][0-9][0-9])\s} $response {\1-} partresponse]} {
+    if {!$nopartresp && $delay > $annoy_partrespevery} {
         incr delay -$annoy_partrespevery
 	set state(ptoid) [thread_after ic $id $annoy_partrespevery \
-		commandresponsedelay_part $delay $final $response $partresponse]
+		commandresponsedelay_part $delay $final $response $f_ix]
 	return
     } elseif {$delay > 0} {
 	set state(ptoid) [thread_after ic $id $delay \
-		commandresponsedelay_after 0 1 $final $response]
+		commandresponsedelay_after $final $response]
 	return
     }
 	
@@ -335,7 +387,7 @@ thread_subproc ic commandresponsedelay {delay nopartresp final response} {
     }
     set state(smtpcmd) {}
     set state(smtpreaderr) {}
-    if {$final} {
+    if {$final>0} {
 	chanset_hide $state(chan) 1 1
 	threadio_puts ic $id $state(chan) $response tellquit_done tellquit_done
     } else {
@@ -344,23 +396,40 @@ thread_subproc ic commandresponsedelay {delay nopartresp final response} {
     }
 }
 
-thread_chainproc ic commandresponsedelay_part {delay final response partresponse} {
+thread_chainproc ic commandresponsedelay_part {delay final response f_ix} {
+    global canonical_hostname fill_msgs
+
+    if {[regexp {^([0-9][0-9][0-9]\-[^\n]*)\n(.*)$} $response dummy \
+	    thisline remainder]} {
+    } elseif {[regexp {^([0-9][0-9][0-9])\s([^\n]*)} $response dummy \
+	    code rhs]} {
+	set thisline "$code-$rhs"
+	set remainder "$code "
+	if {$final!=0} { append remainder "$canonical_hostname " }
+	append remainder [lindex $fill_msgs $f_ix]
+	set f_ix [expr {($f_ix+1) % [llength $fill_msgs]}]
+    } else {
+	error "incomprehensible commandresponsedelay_part response $response"
+    }
     unset state(ptoid)
-    threadio_puts ic $id $state(chan) "$partresponse\n" \
-	    commandresponsedelay_ok commandresponsedelay_err $delay 0 $final $response
+    threadio_puts ic $id $state(chan) "$thisline\n" \
+	    commandresponsedelay_ok commandresponsedelay_err \
+	    $delay 0 $final $remainder $f_ix
 }
 
-thread_chainproc ic commandresponsedelay_err {delay nopartresp final response emsg} {
+thread_chainproc ic commandresponsedelay_err \
+	{delay nopartresp final response f_ix emsg} {
     ic_command_err $emsg
 }
 
-thread_chainproc ic commandresponsedelay_ok {delay nopartresp final response} {
-    ic_commandresponsedelay $delay $nopartresp $final $response
+thread_chainproc ic commandresponsedelay_ok \
+	{delay nopartresp final response f_ix} {
+    ic_commandresponsedelay $delay $nopartresp $final $response $f_ix
 }
 
-thread_chainproc ic commandresponsedelay_after {delay nopartresp final response} {
+thread_chainproc ic commandresponsedelay_after {final response} {
     unset state(ptoid)
-    ic_commandresponsedelay $delay $nopartresp $final $response
+    ic_commandresponsedelay 0 1 $final $response 0
 }
 
 thread_chainproc ic tellquit_done {args} {
@@ -368,6 +437,7 @@ thread_chainproc ic tellquit_done {args} {
 }
 
 thread_subproc ic commandfinalresponse {evtype message} {
+    ic_kill_subthreads
     if {[info exists state(mtachan)]} {
 	threadio_commandresponse ic $id $state(mtachan) quit \
 		{} mtaquit_done mtaquit_done $evtype $message
@@ -417,8 +487,9 @@ thread_subproc ic check_quitting {} {
 }
 
 thread_chainproc ic command_ok {cmd} {
-    global canonical_hostname blacklist_message bland_message admin_chal_timeout
-    global adminsecret blacksite_message allow_saucestate
+    global canonical_hostname blacklist_message bland_message
+    global admin_chal_timeout always_blacklist_site
+    global adminsecret blacksite_message allow_saucestate mixedkinds_message
     after cancel $state(itoid)
     unset state(itoid)
     regexp {^.*} $cmd cmd
@@ -440,7 +511,8 @@ thread_chainproc ic command_ok {cmd} {
 	switch -exact -- $verb {
 	    quit {
 		ic_commandnorhs $rhs
-		ic_commandfinalresponse command "221 $canonical_hostname goodbye"
+		ic_commandfinalresponse command \
+			"221 $canonical_hostname goodbye"
 	    }
 	    helo {
 		ic_helo helo smtp $rhs
@@ -475,7 +547,6 @@ thread_chainproc ic command_ok {cmd} {
 			    ic_mailfrom_ok
 			} elseif {"$as" == "black"} {
 			    set state(mf_message) "You are on the blacklist"
-			    set state(no) "Blacklisted sender `$str'"
 			    ic_mailfrom_ok
 			} elseif {"$as" == "unknown"} {
 			    set state(avfid) [thread_start avf \
@@ -514,54 +585,98 @@ thread_chainproc ic command_ok {cmd} {
 		ic_commandresponse command "252 VRFY not supported by SAUCE."
 	    }
 	    rcpt {
+		incr state(rcptcounter)
 		if {![info exists state(mf_lp)]} {
 		    ic_commandresponse minor "503 need MAIL before RCPT"
 		} elseif {[regexp -nocase -- \
 			{^to:[ \t]*<(.+)@([^@]+)>[ \t]*$} \
 			$rhs all lp dm]} {
+		    set str "$lp@$dm"
 		    if {[catch { address_dequote lp dm } emsg]} {
-			ic_commandresponse major "501 Syntax error in recipient ($emsg)"
+			ic_rcptresponse major badsyntax $str \
+				"501 Syntax error in recipient ($emsg)"
 		    } else {
-			set str "$lp@$dm"
 			set rtcmd "rcpt to:<[lp_quote $lp]@$dm>"
-			set atype [addr_classify $lp $dm]
-			lappend state(att_rcpts) [list $atype $lp@$dm]
-			if {"$atype" == "admin"} {
-			    set state(a_admin) 1
-			    threadio_commandresponse ic $id $state(mtachan) \
-				    $rtcmd {^250} mta_rcptadmin_ok {}
-			} elseif {"$atype" == "bait"} {
-                            set str "Sent mail to bait address $lp@$dm"
-			    set state(add_bl) $str
-			    set state(no) $str
-			    set state(a_normal) 1
-			    ic_commandresponse command "250 $bland_message"
-			} elseif {[info exists state(no)]} {
-			    if {[info exists state(add_bl)]} {
-				set state(a_normal) 1
-				ic_commandresponse command "250 $bland_message"
-			    } else {
-				ic_reject_rcpt $str
+			set atype [addr_classify $lp $dm state]
+			regexp {^[0-9a-z]+} $atype atype_summ
+			set kind $atype
+			switch -glob $atype_summ {
+			    unchecked { set notafter normal }
+			    lax { set notafter {} }
+			    nodelay - normal {
+				set kind normal; set notafter {bait unchecked}
 			    }
-			} elseif {[info exists state(defer)]} {
-			    ic_commandresponse rcpt-defer "450 $state(defer)"
-			} elseif {"[ds_get site-list $state(ra)]" == "black"} {
-			    set state(no) "Blacklisted site \[$state(ra)\]"
-			    ic_reject_rcpt $str
-			} else {
-			    threadio_commandresponse ic $id $state(mtachan) \
-				    $rtcmd {} mta_rcpt_ok {} $atype
+			    bait { set notafter {} }
+			    [45]* {
+				ic_rcptresponse major $atype_summ $str $atype
+				return
+			    }
+			    default { error "internal error - atype $atype" }
 			}
+			set delay_this {}
+			if {"$atype" == "normal" && \
+				[info exists state(delay_msg)]} {
+			    lappend delay_this $state(delay_msg)
+			}
+			if {"[ds_get site-list $state(ra)]" == "black"} {
+			    ic_rej_bl "\[$state(ra)\]" site
+			}
+			set mf $state(mf_lp)@$state(mf_dm)
+			if {"[ds_get addr-list $mf]" == "black"} {
+			    ic_rej_bl $mf "return path"
+			}
+			set rej_this [concat $state(rej_conn) $state(rej_msg)]
+			set defer_this \
+				[concat $state(defer_conn) $state(defer_msg)]
+			switch -exact $atype {
+			    bait {
+				lappend state(add_bl) \
+					"Sent mail to bait address $lp@$dm"
+				if {$always_blacklist_site} ic_blacklist_site
+				ic_rcptresponse command $atype_summ $str \
+					"250 $bland_message"
+				lappend state(a_kinds) $kind
+				return
+			    }
+			    normal - nodelay {
+				foreach {varname code badness} {
+				    rej_this   550 major
+				    defer_this 451 major
+				    delay_this 450 rcpt-defer
+				} {
+				    set resp [join [set $varname] "\n"]
+				    if {[string length $resp]} {
+					smtp_prefix_response $resp $code resp
+					ic_rcptresponse $badness $atype_summ \
+						$str $resp
+					return
+				    }
+				}
+			    }
+			    unchecked - lax {
+			    }
+			    default {
+				error "atype ? $atype"
+			    }
+			}
+			foreach k $notafter {
+			    if {[ic_a_kind $k]} {
+				ic_rcptresponse command $atype_summ $str \
+					"450 $mixedkinds_message"
+				return
+			    }
+			}
+			threadio_commandresponse ic $id $state(mtachan) \
+				$rtcmd {} mta_rcpt_ok {} $kind $atype_summ $str
 		    }
 		} else {
-		    ic_commandresponse major "501 Syntax error in parameter to RCPT"
+		    ic_commandresponse major \
+			    "501 Syntax error in parameter to RCPT"
 		}
 	    }
 	    data {
 		ic_commandnorhs $rhs
-		if {![info exists state(a_admin)] &&
-		    ![info exists state(a_normal)] &&
-		    ![info exists state(a_bypass)]} {
+		if {![llength $state(a_kinds)]} {
 		    ic_commandresponse minor "503 No recipients specified"
 		} else {
 		    threadio_puts ic $id $state(chan) \
@@ -573,21 +688,26 @@ thread_chainproc ic command_ok {cmd} {
 		    set chal [exec -keepnewline \
 			    dd if=/dev/urandom bs=1 count=8 2>/dev/null]
 		    binary scan $chal H* chal
-		    if {[string length $chal] != 16} { error "urandom failed `$chal'" }
+		    if {[string length $chal] != 16} {
+			error "urandom failed `$chal'"
+		    }
 		    append chal [format %08lx [clock seconds]]
 		    set state(lastchal) $chal
 		    ic_commandresponse immed "393 $chal"
 		} elseif {![info exists state(lastchal)]} {
-		    ic_commandresponse major "503 Need SAUCEADMIN on its own first"
+		    ic_commandresponse major \
+			    "503 Need SAUCEADMIN on its own first"
 		} else {
 		    set waschal $state(lastchal)
 		    log notice "$state(desc): ATTEMPTING SWITCH TO ADMIN MODE"
-		    if {![regexp {^([0-9a-f]{16})([0-9a-f]{8})[ \t]+([0-9a-f]{32})$} \
+		    if {![regexp \
+           {^([0-9a-f]{16})([0-9a-f]{8})[ \t]+([0-9a-f]{32})$} \
 			    $rhs all chal wasdate resp]} {
 			ic_commandresponse major "501 \\x{24} \\x{32} please"
 		    } elseif {"$chal$wasdate" != "$waschal"} {
 			ic_commandresponse immed "490 challenge overwritten"
-		    } elseif "[clock seconds] - 0x$wasdate > $admin_chal_timeout" {
+		    } elseif "[clock seconds] - 0x$wasdate \
+			    > $admin_chal_timeout" {
 			ic_commandresponse immed "491 challenge timed out"
 		    } elseif {![string length $adminsecret]} {
 			ic_commandresponse immed "495 admin secret missing"
@@ -630,7 +750,7 @@ thread_chainproc ic command_ok {cmd} {
 		    threadio_commandresponse ic $id $state(mtachan) rset \
 			    {^2[0-9][0-9]} mta_rset_ok {}
 		} else {
-		    catch { unset state(mf_lp) }
+		    ic_msg_resetvars
 		    ic_commandresponse command "250 OK"
 		}
 	    }
@@ -649,13 +769,17 @@ thread_subproc ic helo {helocmd smtpstyle rhs} {
 	ic_commandresponse major "503 HELO or EHLO already specified"
     } elseif {[regexp {^\[(\d+\.\d+\.\d+\.\d+)\]$} $rhs all ipliteral]} {
 	if {$forbid_helo_ipliteral} {
-	    ic_commandresponse major \
-		    "504 IP literal ($rhs) in HELO forbidden by adminstrator"
+	    lappend state(rej_conn) \
+		    "IP literal ($rhs) in HELO forbidden by adminstrator"
+	}
+	ic_find_maildomain $state(rh) $rhs
+    } elseif {![domain_ok $rhs]} {
+	if {[regexp -nocase {[^-_.+@/<>0-9a-z]} $rhs]} {
+	    ic_commandresponse major "501 Invalid characters in HELO domain"
 	} else {
+	    lappend state(rej_conn) "Syntax error in HELO domain `$rhs'"
 	    ic_find_maildomain $state(rh) $rhs
 	}
-    } elseif {![domain_ok $rhs]} {
-	ic_commandresponse major "501 Syntax error in HELO domain"
     } else {
 	if {"[string tolower $rhs]" == "[string tolower $state(rh)]"} {
 	    ic_find_maildomain $rhs $rhs
@@ -664,45 +788,51 @@ thread_subproc ic helo {helocmd smtpstyle rhs} {
 	} elseif {"[ds_get site-list $state(ra)]" == "white"} {
 	    ic_set_maildomain "\[$state(ra)\]" $rhs
 	} else {
-	    set state(dnsid) [thread_start dns "$state(desc) / HELO lookup" $rhs A]
+	    set state(dnsid) [thread_start dns "$state(desc) / HELO lookup" $rhs A 1]
 	    thread_join ic $id dns $state(dnsid) helodns_ok helodns_err $rhs
 	}
     }
 }
 
-thread_chainproc ic helodns_ok {hs answers errors how} {
-    global check_helo_name
+thread_chainproc ic helodns_ok {hs answers emsgstr how} {
+    global check_helo_name require_callingmaildomain_dnsok
     unset state(dnsid)
     if {[llength $answers]} {
 	if {[lsearch -exact $answers $state(ra)] != -1} {
 	    ic_find_maildomain $hs $hs
-	} elseif {$check_helo_name} {
-	    ic_commandresponse major \
-		    "504 HELO name $hs has no address matching $state(ra)"
 	} else {
+	    if {$check_helo_name} {
+		lappend state(rej_conn) \
+			"HELO name $hs has no address matching $state(ra)"
+	    }
 	    ic_find_maildomain $state(rh) $hs
 	}
-    } elseif {$check_helo_name} {
-	ic_commandresponse major \
-		"504 HELO name $hs does not map to any addresses: $errors"
     } else {
+	if {"$how" == "MISCONFIG" && $require_callingmaildomain_dnsok} {
+	    lappend state(rej_conn) \
+		    "HELO name lookup revealed misconfiguration: $emsgstr"
+	} elseif {$check_helo_name} {
+	    lappend state(rej_conn) "HELO name incorrect: $emsgstr"
+	}
 	ic_find_maildomain $state(rh) $hs
     }
 }
 
 thread_chainproc ic helodns_err {hs emsg} {
     unset state(dnsid)
-    ic_commandresponse major "450 HELO name $hs lookup failed: $emsg"
+    lappend state(defer_conn) "HELO name lookup failed: [singleline $emsg]"
+    ic_find_maildomain $state(rh) $hs
 }
 
 thread_subproc ic find_maildomain {chstart hs} {
-    global require_callingmaildomain_name
+    global require_callingmaildomain_name require_reverse_dns
     if {![string match {\[*\]} $chstart]} {
 	ic_findmore_maildomain $chstart $chstart $hs
-    } elseif {$require_callingmaildomain_name} {
-	ic_commandresponse major \
-		"504 Cannot find $state(ra) host name via reverse DNS or HELO"
     } else {
+	if {$require_callingmaildomain_name && !$require_reverse_dns} {
+	    lappend state(rej_conn) \
+		    "Cannot find $state(ra) host name via reverse DNS or HELO"
+	}
 	ic_set_maildomain $chstart $hs
     }
 }
@@ -711,16 +841,21 @@ thread_subproc ic findmore_maildomain {chstart chnow hs} {
     if {[llength [split $chnow .]] == 1} {
 	ic_set_maildomain $chstart $hs
     } else {
-	set state(dnsid) [thread_start dns "$state(desc) / maildomain lookup" $chnow MX]
+	set state(dnsid) [thread_start dns "$state(desc) / maildomain lookup" \
+	                  $chnow MX 0]
 	thread_join ic $id dns $state(dnsid) fch_ok fch_err $chstart $chnow $hs
     }
 }
 
-thread_chainproc ic fch_ok {chstart chnow hs answers errors how} {
+thread_chainproc ic fch_ok {chstart chnow hs answers emsgstr how} {
+    global require_callingmaildomain_dnsok
     unset state(dnsid)
     if {[llength $answers]} {
 	ic_set_maildomain $chnow $hs
     } else {
+	if {"$how" == "MISCONFIG" && $require_callingmaildomain_dnsok} {
+	    lappend state(rej_conn) "While finding mail domain: $emsgstr"
+	}
 	regsub {^[^.]+\.} $chnow {} chnow
 	ic_findmore_maildomain $chstart $chnow $hs
     }
@@ -730,10 +865,10 @@ thread_chainproc ic fch_err {chstart chnow hs emsg} {
     global require_callingmaildomain_dnsok
     unset state(dnsid)
     if {$require_callingmaildomain_dnsok} {
-	ic_commandresponse major "450 Cannot find mail domain (MX for $chnow): $emsg"
-    } else {
-	ic_set_maildomain $chstart $hs
+	lappend state(defer_conn) \
+		"Problem finding mail domain: [singleline $emsg]"
     }
+    ic_set_maildomain $chstart $hs
 }
 
 thread_subproc ic set_maildomain {ch hs} {
@@ -776,9 +911,7 @@ thread_chainproc ic mta_helo_ok {data} {
 	return
     }
     set str "$canonical_hostname hello $state(ident)@$state(rh)"
-    if {![string match {\[*\]} $state(cmdomain)]} {
-	append str " (postmaster@$state(cmdomain)?)"
-    }
+    append str " (postmaster@$state(cmdomain)?)"
     if {"$state(helocmd)" == "helo"} {
 	set op "250 $str"
     } elseif {"$state(helocmd)" == "ehlo"} {
@@ -794,7 +927,6 @@ thread_chainproc ic mta_helo_ok {data} {
 	    }
 	}
 	append op "250 PIPELINING"
-#	append op "250 X-SAUCE"
     } else {
 	error "internal error - ugh? $helocmd"
     }
@@ -817,18 +949,21 @@ thread_subproc ic mailfrom_fail {message} {
 
 thread_chainproc ic mailfrom_avf_ok {ok message} {
     unset state(avfid)
-    if {$ok} {
-	set state(mf_message) $message
-	ic_rbl
-    } else {
-	smtp_prefix_response $message 550 message
-	ic_mailfrom_fail $message
+    if {!$ok} {
+	set sl [singleline "invalid MAIL-FROM: $message"]
+	ic_rej $sl
+	set message "Warning! Rejection likely: $sl"
     }
+    set state(mf_message) $message
+    ic_rbl
 }
 
-thread_chainproc ic mailfrom_avf_err {emsg} {
+thread_chainproc ic mailfrom_avf_err {message} {
     unset state(avfid)
-    ic_mailfrom_fail "450 Unable to verify: [proto_quote $emsg]"
+    set sl [singleline "problematic MAIL-FROM: $message"]
+    lappend state(defer_msg) $sl
+    set state(mf_message) "Warning! Deferral likely: $sl"
+    ic_rbl
 }
 	
 thread_subproc ic rbl {} {
@@ -845,36 +980,49 @@ thread_subproc ic rbl {} {
     foreach rbl $rbls {
 	manyset $rbl dm maa msa rblmsg
 	regsub -all {%d} $rblmsg $dm rblmsg
-	set thread [thread_start dns "$state(desc) / rbl $dm" $tolookup$dm TXT]
+	set thread [thread_start dns "$state(desc) / rbl $dm" $tolookup$dm TXT 1]
 	lappend state(rblids) $thread
-	thread_join ic $id dns $thread rbl_done rbl_err $thread $maa $msa $rblmsg
+	thread_join ic $id dns $thread rbl_done rbl_err \
+		$thread $dm $maa $msa $rblmsg
     }
     ic_rbl_checkdone
 }
 
-thread_chainproc ic rbl_done {thread maa msa rblmsg answers errors etype} {
+thread_chainproc ic rbl_done {thread dm maa msa rblmsg answers emsgstr etype} {
     ic_rbl_rmthread $thread
-    if {[llength $answers]} {
-	set l {}
-	foreach a $answers {
-	    regsub {^\"} $a {} a
-	    regsub {\"$} $a {} a
-	    set a [proto_quote $a]
-	    lappend l $a
+    switch -exact -- $etype {
+	NXDOMAIN - NOTYPE {
 	}
-	regsub -all {%m} $rblmsg [join $l ", "] rblmsg
-	regsub -all {%p} $rblmsg {%} rblmsg
-	if {![string length $maa]} {
-	    set state(no) $rblmsg
-	} else {
-	    ic_rbl_minage addr $state(mf_lp)@$state(mf_dm) $maa $rblmsg
-	    ic_rbl_minage site $state(ra) $msa $rblmsg
+	MISCONFIG {
+	    log notice "RBL misconfigured! $emsgstr"
+	}
+	OK {
+	    if {![llength $answers]} { error "no answers, no error" }
+	    lappend state(rbl_hits) $dm
+	    set l {}
+	    foreach a $answers {
+		regsub {^\"} $a {} a
+		regsub {\"$} $a {} a
+		set a [proto_quote $a]
+		lappend l $a
+	    }
+	    regsub -all {%m} $rblmsg [join $l ", "] rblmsg
+	    regsub -all {%p} $rblmsg {%} rblmsg
+	    if {![string length $maa]} {
+		lappend state(rej_conn) [singleline $rblmsg]
+	    } else {
+		ic_rbl_minage addr $state(mf_lp)@$state(mf_dm) $maa $rblmsg
+		ic_rbl_minage site $state(ra) $msa $rblmsg
+	    }
+	}
+	default {
+	    error "[list $etype $answers $emsgstr] ?"
 	}
     }
     ic_rbl_checkdone
 }
 
-thread_chainproc ic rbl_err {thread maa msa rblmsg emsg} {
+thread_chainproc ic rbl_err {thread dm maa msa rblmsg emsg} {
     ic_rbl_rmthread $thread
     ic_rbl_checkdone
 }
@@ -906,7 +1054,7 @@ thread_subproc ic rbl_minage {what key newminage msg} {
 	debug 2 rbl_minage ... firstcontact $now
     }
     if {$now < $firstcontact+$newminage} {
-	set state(defer) $msg
+	set state(delay_msg) $msg
 	debug 2 rbl_minage ... defer $now $firstcontact+$newminage
     }
 }
@@ -943,35 +1091,23 @@ thread_chainproc ic mta_mailfrom_ok {data} {
     }
 }
 
-thread_chainproc ic mta_rcptadmin_ok {data} {
-    global bland_message
-    set state(a_admin) 1
-    ic_commandresponse command "250 $bland_message"
+thread_subproc ic rcptresponse {evtype atype_summ rcpt response} {
+    regexp {^[0-9][0-9][0-9]} $response code
+    lappend state(att_rcpts) [list $atype_summ $rcpt $code]
+    ic_commandresponse $evtype $response
 }
 
-thread_chainproc ic mta_rcpt_ok {atype data} {
+thread_chainproc ic mta_rcpt_ok {kind atype_summ rcpt data} {
     if {[regexp {^2[0-9][0-9]} $data]} {
-	set state(a_$atype) 1
-	ic_commandresponse command $data
+	lappend state(a_kinds) $kind
+	set evtype command
     } else {
-	ic_commandresponse major $data
+	set evtype major
     }
+    ic_rcptresponse $evtype $atype_summ $rcpt $data
 }
 
 thread_chainproc ic mta_rset_ok {data} {
     catch { unset state(mf_lp) }
     ic_commandresponse command "250 OK"
 }
-
-thread_subproc ic reject_rcpt {str} {
-    global blacklist_message
-    
-    if {[info exists state(add_bl)]} {
-	set msg $blacklist_message
-	set state(whyreject) $state(no)
-    } else {
-	set msg $state(no)
-	set state(whyreject) {}
-    }
-    ic_commandresponse major "550 $msg"
-}
